@@ -3,7 +3,7 @@ package Bot::BasicBot::Pluggable::Module::Nagios;
 use warnings;
 use strict;
 
-our $VERSION = '0.04';
+our $VERSION = '0.05';
 
 use base 'Bot::BasicBot::Pluggable::Module';
 
@@ -80,12 +80,22 @@ sub told {
         }
         my @channels = split /\s+|,/, $channel_list;
         my $instances = $self->get('instances') || [];
-        push @$instances, {
+
+        my $instance = {
             url      => $url,
             user     => $user,
             pass     => $pass,
             channels => \@channels,
         };
+
+        my $poll_result = $self->poll_instance($instance);
+        if ($poll_result->{error}) {
+            my $instance_name = $self->instance_name($instance);
+            return "Failed to poll $instance_name - "
+                . $poll_result->{error} . " - not adding it";
+        }
+
+        push @$instances, $instance;
         $self->set('instances' => $instances);
         return "OK, added Nagios instance to monitor";
     }
@@ -163,15 +173,31 @@ sub told {
             },
             report_statuses => {
                 description => "List of statuses we should notify for"
-                    . " (default: CRITICAL, OK)",
+                    . " (default: CRITICAL, WARNING, OK, UPDATE_FAIL)",
                 validator   => sub {
                     my @statuses = split /[\s,]+/, uc shift;
                     return unless @statuses;
                     return if 
-                        grep { !/^(OK|WARNING|CRITICAL|UNKNOWN)$/ } @statuses;
+                        grep { 
+                            !/^(OK|WARNING|CRITICAL|UNKNOWN|UPDATE_FAIL)$/
+                        } @statuses;
                     return \@statuses;
                 },
             },
+            filter_services => {
+                description => "A regex to match services which"
+                    . " should be ignored",
+                validator => sub {
+                    my $in = shift;
+                    my $re;
+                    eval {
+                        $re = qr/$in/;
+                    };
+                    return if $@;
+                    return $in;
+                },
+            },
+                    
         );
 
         # If we were called without a setting name, reply with the settings
@@ -189,14 +215,16 @@ sub told {
         if (!$validator) {
             return sprintf "Unknown setting '%s' (known settings: %s)",
                 $setting,
-                keys %valid_settings;
+                join ',', keys %valid_settings;
         }
 
         # The validator will return the value (possibly canonicalised) if it was
         # acceptable, or undef if not:
         if (defined(my $valid_value = $validator->($value))) {
             $self->set($setting, $valid_value);
-            return "OK, set $setting to '$valid_value'";
+            my $show_value = ref $valid_value eq 'ARRAY' 
+                ? join(',', @$valid_value) : $valid_value;
+            return "OK, set $setting to '$show_value'";
         } else {
             return "Value '$value' is not valid for setting $setting";
         }
@@ -205,6 +233,7 @@ sub told {
 
 my $last_polled = 0;
 my %last_status;
+my %last_update_failure;
 
 sub tick {
     my ($self) = @_;
@@ -214,52 +243,128 @@ sub tick {
     return if (time - $last_polled < $poll_delay);
     $last_polled = time;
 
+    # OK, time to poll Nagios and report stuff.
+    $self->check_nagios;
+}
+
+sub instance_name {
+    my ($self, $instance) = @_;
+    return join "@", @$instance{qw(user url)};
+}
+
+# Polls the specified Nagios instance and returns all host and service
+# statuses.  Returns a hashref with keys hosts, services, and possibly
+# error, if there was one.
+sub poll_instance {
+    my ($self, $instance) = @_;
+
+    my $ns = Nagios::Scrape->new(
+        username => $instance->{user},
+        password => $instance->{pass},
+        url      => $instance->{url},
+    );
+
+    # Get a list of all hosts, and assemble a lookup hash so we can
+    # easily look up whether a host is down in order to skip reporting
+    # services
+    $ns->host_state(14); # All hosts, including OK ones
+    my @all_hosts;
+    # Nagios::Scrape will die() on error
+    eval { @all_hosts = $ns->get_host_status; 1 };
+
+    if (my $eval_error = $@) {
+        my $error = sprintf "Failed to poll hosts on %s - %s",
+            $self->instance_name($instance), $eval_error;
+        warn $error;
+        return { error => $error };
+    }
+
+    if (!@all_hosts) {
+        my $error = "No hosts returned for " . $self->instance_name($instance);
+        warn $error;
+        return { error => $error };
+    }
+
+    # Get services in all states except PENDING - we want OK ones, too, so
+    # we can easily report problem -> OK transitions
+    # PENDING 1 OK 2 WARNING 4 UNKNOWN 8 CRITICAL 16
+    # 16 + 8 + 4 + 2 = 30 = OK/WARNING/UNKNOWN/CRITICAL
+    # TODO: make state filter configurable
+    $ns->service_state(30);
+
+    my @service_statuses;
+    eval { @service_statuses = $ns->get_service_status; };
+    if (my $eval_error = $@) {
+        my $error = sprintf "Failed to poll services on %s - %s",
+            $self->instance_name($instance), $eval_error;
+        warn $error;
+        return { error => $error };
+    }
+
+    if (!@service_statuses) {
+        my $error = "No services returned for "
+            . $self->instance_name($instance);
+        warn $error;
+        return { error => $error };
+    }
+
+    # OK, looks good, return the service and host statuses we found.
+    return {
+        services => \@service_statuses,
+        hosts    => \@all_hosts,
+    };
+}
+
+sub check_nagios {
+    my ($self) = @_;
+
+    my $repeat_delay = $self->get('repeat_delay') || 15 * 60;
+
+
     # Find out what statuses we should report; do this here, so it's ready for
     # use in the loop later (we don't want to re-do it for every service :) )
     my $report_statuses = $self->get('report_statuses')
-        || [ qw( CRITICAL WARNING OK ) ];
+        || [ qw( CRITICAL WARNING OK UPDATE_FAIL ) ];
     my %should_report = map { $_ => 1 } @$report_statuses;
 
 
     my $instances = $self->get('instances') || [];
     instance:
     for my $instance (@$instances) {
-        my $ns = Nagios::Scrape->new(
-            username => $instance->{user},
-            password => $instance->{pass},
-            url      => $instance->{url},
-        );
+        my $instance_name = $self->instance_name($instance);
+        my $result = $self->poll_instance($instance);
 
-        # Get a list of all hosts, and assemble a lookup hash so we can
-        # easily look up whether a host is down in order to skip reporting
-        # services
-        $ns->host_state(14); # All hosts, including OK ones
-        my @all_hosts = $ns->get_host_status;
+        # First, if there was an error polling this instance, report it and move
+        # on:
+        if ($result->{error} && $should_report{UPDATE_FAIL}) {
+            my $instance_name = $self->instance_name($instance);
+            if (time - $last_update_failure{$instance_name}
+                > $repeat_delay)
+            {
+                for my $channel (@{ $instance->{channels} }) {
+                    $self->tell(
+                        $channel,
+                        "NAGIOS: Update failure for $instance_name: "
+                            . $result->{error}
+                    );
+                }
+            }
+            $last_update_failure{$instance_name} = time;
+            next instance;
+        }
+        
         my %host_down  = 
             map  { $_->{host} => 1        } 
             grep { $_->{status} eq 'DOWN' }
-            @all_hosts;
+            @{ $result->{hosts} };
 
 
-        # Get services in all states except PENDING - we want OK ones, too, so
-        # we can easily report problem -> OK transitions
-        # PENDING 1 OK 2 WARNING 4 UNKNOWN 8 CRITICAL 16
-        # 16 + 8 + 4 + 2 = 30 = OK/WARNING/UNKNOWN/CRITICAL
-        # TODO: make state filter configurable
-        $ns->service_state(30);
-    
-        my @service_statuses = $ns->get_service_status;
-
-
-        # Key to use for this instance in %last_status
-        my $instance_key = join '_', $instance->{url}, $instance->{user};
-        
-        my $instance_statuses = $last_status{$instance_key} ||= {};
+        my $instance_statuses = $last_status{$instance_name} ||= {};
 
         # Firstly, report host status changes:
         my @host_reports;
         host:
-        for my $host (@all_hosts) {
+        for my $host (@{ $result->{hosts} }) {
             if (my $last_status = $instance_statuses->{$host->{host}}) {
                 # If it was UP before and is still UP, move on swiftly
                 next if $last_status->{status} eq 'UP'
@@ -272,8 +377,7 @@ sub tick {
                     next host;
                 }
 
-                # OK, announce that this host is down, and remember that did
-                # so
+                # OK, we need to announce that this host is down
                 push @host_reports, $host;
             } else {
                 # We've not seen this one before; if it's 'UP', just remember 
@@ -302,16 +406,20 @@ sub tick {
             }
         }
 
-
-
+        # Fetch & compile the service filtering regex once, rather than once for
+        # each service:
+        my $filter_re = $self->get('filter_services') || '';
+        $filter_re = qr($filter_re) if $filter_re;
 
         # Group problems by host, ignoring any which we've already reported 
         # recently, or which are on a host which is down.
         my %service_by_host;
         service:
-        for my $service (@service_statuses) {
+        for my $service (@{ $result->{services} }) {
             next if $host_down{$service->{host}};
 
+            # Skip it if we should be filtering it out
+            next if $filter_re && $service->{service} !~ $filter_re;
 
             # See how many check attempts have found the service in this status;
             # if it's not enough for Nagios to send alerts, don't alert on IRC.
@@ -376,6 +484,15 @@ Plenty of improvements are planned, including:
 
 =over 4
 
+=item * Better documentation
+
+I need to improve the module's documentation.  For now, extra information is
+available by saying <help nagios> to the bot on IRC, or C<nagios set> with no
+setting name for a list of valid settings with descriptions.
+
+I'd rather work out a good way to auto-generate documentation from the settings
+definitions in the code in order to make sure the docs stay in sync.
+
 =item * Acknowledging problems
 
 It should probably be possible to acknowledge a reported problem, preventing
@@ -458,7 +575,7 @@ L<http://search.cpan.org/dist/Bot-BasicBot-Pluggable-Module-Nagios/>
 
 =head1 LICENSE AND COPYRIGHT
 
-Copyright 2011 David Precious.
+Copyright 2011-2017 David Precious.
 
 This program is free software; you can redistribute it and/or modify it
 under the terms of either: the GNU General Public License as published
